#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "RC_Blanko.h"

static uint8_t new;
static uint8_t cmd;
static uint8_t lasttime;
static uint8_t bitpos;
static uint8_t threshold;
static uint8_t checktime;


void RC_Init(void)
{
  MCUCR = 1<<ISC11 | 1<<ISC10; // trigger on rising edge
  GIMSK = 1<<INT1;  // enable INT1
}


uint8_t RC_PollCmd(void)
{
  uint8_t result = 0;
  if(new)
  {
    result = cmd;
    new = 0;
  }
  return result;
}



SIGNAL(TIMER1_OVF_vect)
{
  if(lasttime == checktime)
  {
    bitpos = 0;
    TIMSK = 0;         //disable timer interrupt
    GIMSK = 1<<INT1;   //enable INT1
  }
  else
    checktime = lasttime;
}


SIGNAL(INT1_vect)
{
  uint8_t time;
  uint8_t bitpos_l;
  time = (TCNT1 >> 8);
  bitpos_l = bitpos;

  if(bitpos_l & 0x80)
  {
    uint8_t diff;
    diff = (time - lasttime);
    bitpos_l &= 0x7F;
    if(bitpos_l == 0)
    {
      threshold = diff>>2;
    }

    if(bitpos_l == 20)
    {
      new = 1;
      GIMSK = 0;  // disable INT1
    }

    cmd <<= 1;
    bitpos_l++;

    if(diff > threshold) //3500
    {
      cmd++;
      bitpos_l |= 0x80;
    }
  }
  else
  {
    bitpos_l |= 0x80;
  }
  bitpos = bitpos_l;
  lasttime = time;

  /* restart timeout */
  TIFR = 1<<TOV1;      //clear flag
  TIMSK = 1<<TOIE1;     //enable timer interrupt
}

