#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <inttypes.h>
#include <util/delay.h>

#include "RC_Blanko.h"
#include "PWM16_RGB.h"


typedef struct {
  uint16_t H;
  uint8_t  S;
  uint8_t  V;
} hsv_t;

typedef hsv_t (*RCcb_t)(hsv_t);

typedef struct {
  uint8_t cmd;
  RCcb_t  Callback;
} rccmd_t;

hsv_t CMD_On(hsv_t Color);
hsv_t CMD_Off(hsv_t Color);
hsv_t CMD_Fade(hsv_t Color);
hsv_t CMD_BrightUp(hsv_t Color);
hsv_t CMD_BrightDown(hsv_t Color);
hsv_t CMD_Smooth(hsv_t Color);

hsv_t CMD_Orange(hsv_t Color);
hsv_t CMD_Strobe(hsv_t Color);
hsv_t CMD_DarkOrange(hsv_t Color);
hsv_t CMD_Green(hsv_t Color);
hsv_t CMD_Blue(hsv_t Color);
hsv_t CMD_LightBlue(hsv_t Color);
hsv_t CMD_Purple(hsv_t Color);
hsv_t CMD_DarkPink(hsv_t Color);
hsv_t CMD_GreenBlue(hsv_t Color);
hsv_t CMD_Cyan(hsv_t Color);
hsv_t CMD_Yellow(hsv_t Color);
hsv_t CMD_FullWhite(hsv_t Color);
hsv_t CMD_FullBlue(hsv_t Color);
hsv_t CMD_Flash(hsv_t Color);
hsv_t CMD_FullRed(hsv_t Color);
hsv_t CMD_FullGreen(hsv_t Color);
hsv_t CMD_Red(hsv_t Color);
hsv_t CMD_Pink(hsv_t Color);

const PROGMEM rccmd_t RC_Dispatch[] = {
  {BLANKO_ORANGE, CMD_Orange},
  {BLANKO_STROBE, CMD_Strobe},
  {BLANKO_DARKORANGE, CMD_DarkOrange},
  {BLANKO_FADE, CMD_Fade},
  {BLANKO_GREEN, CMD_Green},
  {BLANKO_BLUE, CMD_Blue},
  {BLANKO_LIGHTBLUE, CMD_LightBlue},
  {BLANKO_PURPLE, CMD_Purple},
  {BLANKO_DARKPINK, CMD_DarkPink},
  {BLANKO_SMOOTH, CMD_Smooth},
  {BLANKO_GREENBLUE, CMD_GreenBlue},
  {BLANKO_CYAN, CMD_Cyan},
  {BLANKO_YELLOW, CMD_Yellow},
  {BLANKO_FULL_WHITE, CMD_FullWhite},
  {BLANKO_FULL_BLUE, CMD_FullBlue},
  {BLANKO_ON, CMD_On},
  {BLANKO_BRIGHT_UP, CMD_BrightUp},
  {BLANKO_FLASH, CMD_Flash},
  {BLANKO_BRIGHT_DOWN, CMD_BrightDown},
  {BLANKO_FULL_RED, CMD_FullRed},
  {BLANKO_FULL_GREEN, CMD_FullGreen},
  {BLANKO_RED, CMD_Red},
  {BLANKO_PINK, CMD_Pink},
  {BLANKO_OFF, CMD_Off},
};

void hsv_to_rgb (uint16_t h, uint8_t s, uint8_t v)
{
#if 1
  uint8_t r, g, b;
  if(s)
  {
    uint8_t i, f, q;

    i = h>>8;
    f = h;
    if(i & 0x01)
      f = 255 - f;
    q = ((v + 1) * f) >> 8;

    switch(i)
    {
    default:
    case 0:
      r = v; g = q; b = 0;
      break;
    case 1:
      r = q; g = v; b = 0;
      break;
    case 2:
      r = 0; g = v; b = q;
      break;
    case 3:
      r = 0; g = q; b = v;
      break;
    case 4:
      r = q; g = 0; b = v;
      break;
    case 5:
      r = v; g = 0; b = q;
      break;
    }
  }
  else
  {
    r = g = b = v;
  }
  PWMRGB_Update(r, g, b);
#endif
}


uint16_t lsfr(uint16_t seed, uint16_t poly)
{
  asm volatile ("lsl  %A[seed]          \n\t"
                "rol  %B[seed]          \n\t"
                "brcc 1f               \n\t"
                "eor  %A[seed],%A[poly]  \n\t"
                "eor  %B[seed],%B[poly]  \n\t"
                "1:                    "
                : "=r" (seed)
                : [seed] "r" (seed), [poly] "r" (poly)
  );
  return seed;
}


void* Dispatch(uint8_t i, void *pD, uint8_t cmd);


uint8_t EEPROM_RingInit(uint8_t size)
{
  uint8_t CurAdr, NextAdr, CurKey, NextKey;
  CurAdr = 0;
  CurKey = eeprom_read_byte(CurAdr);
  size++;
  for(;;)
  {
    NextAdr = CurAdr + size;
    if(NextAdr > (E2END - size))
      return CurAdr;
    NextKey = eeprom_read_byte(NextAdr);
    if((NextKey - CurKey) != 1)
      return CurAdr;
    CurAdr = NextAdr;
    CurKey = NextKey;
  }
}

void EEPROM_RingRead(uint8_t adr, uint8_t *p, uint8_t size)
{
  eeprom_read_block(p, adr + 1, size);
}

void EEPROM_RingWrite(uint8_t adr, uint8_t *p, uint8_t size)
{
  uint8_t tmp;
  tmp = eeprom_read_byte(adr);
  adr += (size + 1);
  if((adr + size + 1) > E2END)
    adr = 0;
  tmp++;
  eeprom_write_block(p, adr + 1, size);
  eeprom_write_byte(adr, tmp);
}


void __attribute__((OS_main))
main(void)
{
  uint8_t Cmd;
  hsv_t Color;
  uint16_t rand;
  uint8_t adr;
  //Debug_Init();
  PWMRGB_Init();
  RC_Init();
  adr = EEPROM_RingInit(2);
  EEPROM_RingRead(adr, (uint8_t*) &rand, 2);
//  if(!rand)
//    rand = 0x124A;
  rand = lsfr(rand, 0x100B);

  Color.H = rand;
  Color.S = 255;
  Color.V = 255;              // HSV Startwerte: rot, volle Helligkeit

  EEPROM_RingWrite(adr, (uint8_t*)&rand, 2);

#if 0
  eeprom_read_block(buf, 0, sizeof(buf));
  DH("EEPROM1:", sizeof(buf), buf);
  eeprom_read_block(buf, sizeof(buf), sizeof(buf));
  DH("EEPROM2:", sizeof(buf), buf);
  //eeprom_write_word(&seed, rand);
  while(1);
#endif

  sei();
  for(;;)
  {
#if 1
    Color.H %= 0x0600;
#else
    Color.H = (((Color.H>>8) * 6)>>8) | (Color.H & 0xFF);
#endif
    hsv_to_rgb(Color.H, Color.S, Color.V);        // HSV to RGB

    Cmd = RC_PollCmd();
    if(Cmd)
    {
      RCcb_t Callback;
      Callback = Dispatch(sizeof(RC_Dispatch)/sizeof(rccmd_t), RC_Dispatch, Cmd);
      if(Callback)
      {
        Color = Callback(Color);
      }
    }
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_mode();                   /* start sleeping here */

    if(SREG & (1<<SREG_T))
    {
      asm volatile ("clt");
      Color.H++;
    }
  }
}


hsv_t CMD_BrightUp(hsv_t Color)
{
  if(Color.V < 255)
    Color.V += 5;
  return Color;
}

hsv_t CMD_BrightDown(hsv_t Color)
{
  if(Color.V > 100)
    Color.V -= 5;
  return Color;
}

hsv_t CMD_Off(hsv_t Color)
{
  WDTCSR = 0;
  Color.V = 0;
  return Color;
}

hsv_t CMD_On(hsv_t Color)
{
  Color.V = 255;
  return Color;
}

hsv_t CMD_Fade(hsv_t Color)
{
  if(WDTCSR & (1<<WDIE))
    WDTCSR = 0;
  else
    WDTCSR = 1<<WDIE;
  return Color;
}

hsv_t CMD_Smooth(hsv_t Color)
{
  if(WDTCSR & (1<<WDIE))
    WDTCSR = 0;
  else
    WDTCSR = 1<<WDIE | 1<<WDP1;
  return Color;
}

hsv_t CMD_Flash(hsv_t Color)
{
  return Color;
}

hsv_t CMD_Strobe(hsv_t Color)
{
  return Color;
}

hsv_t CMD_FullWhite(hsv_t Color)
{
  Color.S = 0;
  return Color;
}


hsv_t CMD_FullBlue(hsv_t Color)
{
  Color.H = 0x0400 | 0;
  return Color;
}

hsv_t CMD_Blue(hsv_t Color)
{
  Color.H = 0x0400 | 64;
  return Color;
}

hsv_t CMD_Purple(hsv_t Color)
{
  Color.H = 0x0400 | 128;
  return Color;
}

hsv_t CMD_DarkPink(hsv_t Color)
{
  Color.H = 0x0400 | 192;
  return Color;
}

hsv_t CMD_Pink(hsv_t Color)
{
  Color.H = 0x0500 | 0;
  return Color;
}


hsv_t CMD_FullGreen(hsv_t Color)
{
  Color.H = 0x0200 | 0;
  return Color;
}

hsv_t CMD_Green(hsv_t Color)
{
  Color.H = 0x0200 | 64;
  return Color;
}

hsv_t CMD_GreenBlue(hsv_t Color)
{
  Color.H = 0x0200 | 128;
  return Color;
}

hsv_t CMD_Cyan(hsv_t Color)
{
  Color.H = 0x0200 | 192;
  return Color;
}

hsv_t CMD_LightBlue(hsv_t Color)
{
  Color.H = 0x0300 | 0;
  return Color;
}



hsv_t CMD_FullRed(hsv_t Color)
{
  Color.H = 0x0000 | 0;
  return Color;
}

hsv_t CMD_Red(hsv_t Color)
{
  Color.H = 0x0000 | 64;
  return Color;
}

hsv_t CMD_DarkOrange(hsv_t Color)
{
  Color.H = 0x0000 | 128;
  return Color;
}

hsv_t CMD_Orange(hsv_t Color)
{
  Color.H = 0x0000 | 192;
  return Color;
}

hsv_t CMD_Yellow(hsv_t Color)
{
  Color.H = 0x0100 | 0;
  return Color;
}


void* Dispatch(uint8_t i, void *pD, uint8_t cmd)
{
  while(i--)
  {
    if(pgm_read_byte(&pD[0]) == cmd)
    {
      pD = (void*) pgm_read_word(&pD[1]);
      return pD;
    }
    pD += 3;
  }
  return 0;
}

ISR(WDT_OVERFLOW_vect, ISR_NAKED)
{
  asm volatile (
      "set"                  "\n\t"
      "reti"
  );
}


#define RING(type, size)  struct { uint8_t read; volatile uint8_t write; type data[size];}

#define RING_GetRead(p)      &(p)->data[(p)->read], (p)->read++, (p)->read %= (sizeof((p)->data)/sizeof((p)->data[0]))
#define RING_GetWrite(p)     &(p)->data[(p)->write], (p)->write++, (p)->write %= (sizeof((p)->data)/sizeof((p)->data[0]))
#define RING_IsNotEmpty(p)   ((p)->read != (p)->write)
